'use strict';

const TOKEN = {
  STYLE_BLOCK_START: "{",
  STYLE_NAME_DELIMITER: " ",
  STYLE_BLOCK_END: "}",
  ESCAPE_CHAR: "\\"
};

const STATE = {
  INIT: Symbol("INIT"),
  TEXT: Symbol("TEXT"),
  IGNORE: Symbol("IGNORE"),
  STYLE_BLOCK_ID: Symbol("STYLE_BLOCK_ID"),
  STYLE_BLOCK_TEXT: Symbol("STYLE_BLOCK_TEXT"),
  STYLE_BLOCK_START: Symbol("STYLE_BLOCK_START"),
  STYLE_BLOCK_END: Symbol("STYLE_BLOCK_END")
};

const TYPE = {
  END_OF_FILE: "END_OF_FILE",
  STYLE_BLOCK_START: "STYLE_BLOCK_START",
  STYLE_BLOCK_ID: "STYLE_BLOCK_ID",
  STYLE_BLOCK_TEXT: "STYLE_BLOCK_TEXT",
  STYLE_BLOCK_END: "STYLE_BLOCK_END",
  TEXT: "TEXT"
};

function* scanner(input) {
  let i = 0;
  let state = STATE.INIT;
  let lex = "";

  while (true) {
    const token = input[i++];

    if (token === undefined) {
      if (state === STATE.TEXT && lex) {
        yield { type: "TEXT", lex };
      }

      return { type: TYPE.END_OF_FILE };
    }

    switch (state) {
      case STATE.INIT:
        {
          if (token === TOKEN.STYLE_BLOCK_START) {
            state = STATE.STYLE_BLOCK_ID;
            yield { type: TYPE.STYLE_BLOCK_START };
            continue;
          } else if (token === TOKEN.ESCAPE_CHAR) {
            state = STATE.IGNORE;
            continue;
          } else {
            state = STATE.TEXT;
            lex += token;
            continue;
          }
        }
      case STATE.TEXT:
        {
          if (token === TOKEN.STYLE_BLOCK_START) {
            state = STATE.STYLE_BLOCK_ID;

            yield { type: TYPE.TEXT, lex };
            yield { type: TYPE.STYLE_BLOCK_START };

            lex = "";
            continue;
          }

          if (token === TOKEN.ESCAPE_CHAR) {
            state = STATE.IGNORE;
            continue;
          }

          lex += token;
          continue;
        }
      case STATE.IGNORE:
        {
          state = STATE.TEXT;
          lex += token;
          continue;
        }
      case STATE.STYLE_BLOCK_ID:
        {
          if (token === TOKEN.STYLE_NAME_DELIMITER) {
            state = STATE.STYLE_BLOCK_TEXT;

            yield { type: TYPE.STYLE_BLOCK_ID, lex };

            lex = "";
            continue;
          }

          if (token === TOKEN.STYLE_BLOCK_END) {
            state = STATE.TEXT;

            yield { type: TYPE.STYLE_BLOCK_ID, lex };
            yield { type: TYPE.STYLE_BLOCK_END };

            lex = "";
            continue;
          }

          if (token.match(/\s/)) {
            state = STATE.STYLE_BLOCK_TEXT;

            yield { type: TYPE.STYLE_BLOCK_ID, lex };

            lex = "";
            continue;
          }

          lex += token;
          continue;
        }
      case STATE.STYLE_BLOCK_TEXT:
        {
          if (token === TOKEN.STYLE_BLOCK_END) {
            state = STATE.INIT;

            yield { type: TYPE.STYLE_BLOCK_TEXT, lex };
            yield { type: TYPE.STYLE_BLOCK_END };

            lex = "";
            continue;
          }

          lex += token;
          continue;
        }
    }
  }
}

const TYPE$1 = {
  TEXT: "TEXT",
  STYLED: "STYLED"
};

let stack = null;
let tokenGenerator = null;

function createText(value) {
  stack.push({
    type: TYPE$1.TEXT,
    value
  });
}

function createStyled({ id, value = null }) {
  stack.push({
    type: TYPE$1.STYLED,
    value,
    id
  });
}

function parser(input) {
  stack = [];
  tokenGenerator = scanner(input);

  return Init();
}

function Init() {
  const {
    value: { type, lex }
  } = tokenGenerator.next();

  if (type === TYPE.TEXT) {
    createText(lex);

    return Text();
  }

  if (type === TYPE.STYLE_BLOCK_START) {
    return StyledID();
  }

  throw new SyntaxError(`Unexpected token: ${type}`);
}

function StyledID() {
  const {
    value: { type, lex: id },
    done
  } = tokenGenerator.next();

  if (type === TYPE.STYLE_BLOCK_ID) {
    if (!id) {
      throw new SyntaxError(`Unexpected token: ${type}. Token has to have lex.`);
    }

    return StyledText({
      id
    });
  }

  throw new SyntaxError(`Unexpected token: ${type}`);
}

function StyledText(styledObject) {
  const {
    value: { type, lex: value },
    done
  } = tokenGenerator.next();

  if (type === TYPE.STYLE_BLOCK_TEXT) {
    return StyledEnd(Object.assign({
      value
    }, styledObject));
  }

  if (type === TYPE.STYLE_BLOCK_END) {
    createStyled(Object.assign({}, styledObject));

    return Text();
  }

  throw new SyntaxError(`Unexpected token: ${type}`);
}

function StyledEnd(styledObject) {
  const {
    value: { type, lex },
    done
  } = tokenGenerator.next();
  if (type === TYPE.STYLE_BLOCK_END) {
    createStyled(Object.assign({}, styledObject));
    return Text();
  }

  throw new SyntaxError(`Unexpected token: ${type}`);
}

function Text() {
  const {
    value: { type, lex },
    done
  } = tokenGenerator.next();

  if (type === TYPE.END_OF_FILE) {
    return stack;
  }

  if (type === TYPE.STYLE_BLOCK_START) {
    return StyledID();
  }

  if (type === TYPE.TEXT) {
    createText(lex);

    return Text();
  }

  throw new SyntaxError(`Unexpected token: ${type}`);
}

function convertStyleObjectToString(styleObject) {
  return Object.entries(styleObject).map(([rule, value]) => {
    return `${rule.split(/(?=[A-Z])/).join("-").toLowerCase()}:${value}`;
  }).join(";");
}

function styled(recipe, stylesDictionary = null) {
  const parsedRecipe = parser(recipe);
  const [text, styles] = parsedRecipe.reduce((config, currentValue, currentIndex, array) => {
    const { type, value, id } = currentValue;

    if (!stylesDictionary) {
      config[0] += value;
      return config;
    }

    config[0] += `%c${value || " "}`;

    let style = "";

    if (type === TYPE$1.STYLED) {
      const styleObject = stylesDictionary[id];
      if (!styleObject) {
        throw new Error(`Style for "${id}" doesn't found`);
      }

      style = convertStyleObjectToString(styleObject);
    }

    config[1].push(style);
    return config;
  }, ["", []]);

  console.log(text, ...styles);
}

module.exports = styled;
